import os
import tempfile
import shutil
import pyfastaq
from iva import edge, common

class Error (Exception): pass

HIT_AT_START = 0
HIT_AT_END = 1
HIT_AT_BOTH_ENDS = 2
HIT_NO_ENDS = 3

def run_nucmer(query, ref, outfile, min_id=95, min_length=100, breaklen=200):
    query = os.path.abspath(query)
    ref = os.path.abspath(ref)
    outfile = os.path.abspath(outfile)
    tmpdir = tempfile.mkdtemp(prefix='tmp.run_nucmer.', dir=os.getcwd())
    original_dir = os.getcwd()
    os.chdir(tmpdir)
    script = 'run_nucmer.sh'
    f = pyfastaq.utils.open_file_write(script)
    print('nucmer --maxmatch -p p -b', breaklen, ref, query, file=f)
    print('delta-filter -i', min_id, '-l', min_length, 'p.delta > p.delta.filter', file=f)
    print('show-coords -dTlro p.delta.filter >', outfile, file=f)
    pyfastaq.utils.close(f)
    common.syscall('bash ' + script)
    os.chdir(original_dir)
    shutil.rmtree(tmpdir)


def file_reader(fname):
    f = pyfastaq.utils.open_file_read(fname)
    in_header = True

    for line in f:
        if in_header:
            if line.startswith('['):
                in_header = False
            continue
        yield NucmerHit(line)

    pyfastaq.utils.close(f)


class NucmerHit:
    def __init__(self, line):
        # [S1]  [E1]    [S2]    [E2]    [LEN 1] [LEN 2] [% IDY] [LEN R] [LEN Q] [FRM]   [TAGS]
        #1162    25768   24536   4   24607   24533   99.32   640851  24536   1   -1  MAL1    NODE_25757_length_24482_cov_18.920391   [CONTAINS]

        try:
            l = line.rstrip().split('\t')
            self.ref_start = int(l[0]) - 1
            self.ref_end = int(l[1]) - 1
            self.qry_start = int(l[2]) - 1
            self.qry_end = int(l[3]) - 1
            self.hit_length_ref = int(l[4])
            self.hit_length_qry = int(l[5])
            self.percent_identity = float(l[6])
            self.ref_length = int(l[7])
            self.qry_length = int(l[8])
            self.frame = int(l[9])
            self.ref_name = l[11]
            self.qry_name = l[12]

        except:
            raise Error('Error reading this nucmer line:\n' + line)


    def __eq__(self, other):
        return type(other) is type(self) and self.__dict__ == other.__dict__


    def __hash__(self):
        return hash((self.ref_start, self.ref_end, self.qry_start, self.qry_end, self.hit_length_ref, self.hit_length_qry, self.percent_identity, self.ref_length, self.qry_length, self.frame, self.ref_name, self.qry_name))


    def _swap(self):
        self.ref_start, self.qry_start = self.qry_start, self.ref_start
        self.ref_end, self.qry_end = self.qry_end, self.ref_end
        self.hit_length_ref, self.hit_length_qry = self.hit_length_qry, self.hit_length_ref
        self.ref_length, self.qry_length = self.qry_length, self.ref_length
        self.ref_name, self.qry_name = self.qry_name, self.ref_name


    def qry_coords(self):
        return pyfastaq.intervals.Interval(min(self.qry_start, self.qry_end), max(self.qry_start, self.qry_end))


    def ref_coords(self):
        return pyfastaq.intervals.Interval(min(self.ref_start, self.ref_end), max(self.ref_start, self.ref_end))


    def on_same_strand(self):
        return (self.ref_start < self.ref_end) == (self.qry_start < self.qry_end)


    def sort(self):
        if self.ref_name > self.qry_name:
            self._swap()

        if self.ref_start > self.ref_end:
            self.ref_start, self.ref_end = self.ref_end, self.ref_start
            self.qry_start, self.qry_end = self.qry_end, self.qry_start

    def is_self_hit(self):
        return self.ref_name == self.qry_name \
                and self.ref_start == self.qry_start \
                and self.ref_end == self.qry_end \
                and self.percent_identity == 100


    def to_graph_edge(self, min_overlap_length=200, end_tolerance=50, min_identity=99):
        if self.percent_identity < min_identity or min(self.hit_length_qry, self.hit_length_ref) < min_overlap_length:
            return None
        ref_at_ends = self._is_at_ends(tolerance=end_tolerance)
        qry_at_ends = self._is_at_ends(use_qry=True, tolerance=end_tolerance)
        bad_values = set([HIT_NO_ENDS, HIT_AT_BOTH_ENDS])
        if ref_at_ends in bad_values or qry_at_ends in bad_values:
            return None

        good_values = set([HIT_AT_START, HIT_AT_END])
        assert qry_at_ends in good_values
        assert ref_at_ends in good_values
        qry_fwd_strand = self.qry_start < self.qry_end
        ref_fwd_strand = self.ref_start < self.ref_end

        if ((qry_fwd_strand == ref_fwd_strand) == (qry_at_ends == ref_at_ends)):
            return None
        elif qry_fwd_strand:
            if ref_fwd_strand:
                if qry_at_ends == HIT_AT_END:
                    assert ref_at_ends == HIT_AT_START
                    qry_first = True
                elif qry_at_ends == HIT_AT_START:
                    assert ref_at_ends == HIT_AT_END
                    qry_first = False
                else:
                    raise Error('Error in mummer.to_graph_edge(). Cannot continue')
            else:
                if qry_at_ends == HIT_AT_END:
                    assert ref_at_ends == HIT_AT_END
                    qry_first = True
                elif qry_at_ends == HIT_AT_START:
                    assert ref_at_ends == HIT_AT_START
                    qry_first = False
                else:
                    raise Error('Error in mummer.to_graph_edge(). Cannot continue')
        else:
            if ref_fwd_strand:
                if qry_at_ends == HIT_AT_END:
                    assert ref_at_ends == HIT_AT_END
                    qry_first = False
                elif qry_at_ends == HIT_AT_START:
                    assert ref_at_ends == HIT_AT_START
                    qry_first = True
                else:
                    raise Error('Error in mummer.to_graph_edge(). Cannot continue')
            else:
                if qry_at_ends == HIT_AT_END:
                    assert ref_at_ends == HIT_AT_START
                    qry_first = False
                elif qry_at_ends == HIT_AT_START:
                    assert ref_at_ends == HIT_AT_END
                    qry_first = True
                else:
                    raise Error('Error in mummer.to_graph_edge(). Cannot continue')
        if qry_first:
            return edge.Edge(self.qry_name, self.qry_start, self.qry_end, self.ref_name, self.ref_start, self.ref_end)
        else:
            return edge.Edge(self.ref_name, self.ref_start, self.ref_end, self.qry_name, self.qry_start, self.qry_end)


    def _is_at_ends(self, use_qry=False, tolerance=50):
        if use_qry:
            at_start = min(self.qry_start, self.qry_end) < tolerance
            at_end = self.qry_length - 1 - max(self.qry_start, self.qry_end) <= tolerance
        else:
            at_start = min(self.ref_start, self.ref_end) < tolerance
            at_end = self.ref_length - 1 - max(self.ref_start, self.ref_end) <= tolerance

        if at_start and at_end:
            return HIT_AT_BOTH_ENDS
        elif at_start:
            return HIT_AT_START
        elif at_end:
            return HIT_AT_END
        else:
            return HIT_NO_ENDS


    def __str__(self):
        return '\t'.join(str(x) for x in
            [self.ref_start,
            self.ref_end,
            self.qry_start,
            self.qry_end,
            self.hit_length_ref,
            self.hit_length_qry,
            '{0:.2f}'.format(self.percent_identity),
            self.ref_length,
            self.qry_length,
            self.frame,
            self.ref_name,
            self.qry_name])




import copy
from pyfastaq import intervals

class Error (Exception): pass

LEFT = 0
RIGHT = 1

class Edge:
    def __init__(self, contig1, start1, end1, contig2, start2, end2):
        self.names = [contig1, contig2]
        self.rev = {
            contig1: start1 > end1,
            contig2: start2 > end2
        }
        self.coords = {
            contig1: intervals.Interval(min(start1, end1), max(start1, end1)),
            contig2: intervals.Interval(min(start2, end2), max(start2, end2))
        }

        if self.names[0] > self.names[1]:
            self.reverse()


    def __str__(self):
        contig1 = self.names[0]
        contig2 = self.names[1]
        return contig1 + ',' + str(self.coords[contig1]) + ',' + str(self.rev[contig1]) + '; ' + \
               contig2 + ',' + str(self.coords[contig2]) + ',' + str(self.rev[contig2])


    def __eq__(self, other):
        if set(self.names) != set(other.names):
            return False
        if self.names[0] != other.names[0]:
            other.reverse()
        equal = type(other) is type(self) and self.__dict__ == other.__dict__
        if self.names[0] != other.names[0]:
            other.reverse()
        return equal


    def open_end(self, contig):
        assert contig in self.coords
        is_rev = self.rev[contig]
        is_first = (contig == self.names[0])
        if is_first:
            if self.rev[contig]:
                return RIGHT
            else:
                return LEFT
        else:
            if self.rev[contig]:
                return LEFT
            else:
                return RIGHT


    def reverse(self):
        self.names = self.names[::-1]
        self.rev[self.names[0]] = not self.rev[self.names[0]]
        self.rev[self.names[1]] = not self.rev[self.names[1]]


    def _make_contig_forwards(self, contig):
        assert contig in self.names
        if self.rev[contig]:
            self.reverse()


    def make_contig_first(self, contig):
        if contig != self.names[0]:
            assert contig == self.names[1]
            self.reverse()


    def _change_hit_coords_with_intersection(self, contig, new_interval):
        assert contig in self.names
        coords_intersection = self.coords[contig].intersection(new_interval)
        if coords_intersection is None:
            return False
        other_contig = [x for x in self.names if x != contig][0]
        left_bases_lost = coords_intersection.start - self.coords[contig].start
        right_bases_lost = self.coords[contig].end - coords_intersection.end
        self.coords[contig] = coords_intersection
        if self.rev[contig] != self.rev[other_contig]:
            left_bases_lost, right_bases_lost = right_bases_lost, left_bases_lost

        self.coords[other_contig].start += left_bases_lost
        self.coords[other_contig].end -= right_bases_lost
        return True


    def merge_into(self, other_edge, contig):
        other_edge = copy.copy(other_edge)
        assert contig in self.names
        assert contig in other_edge.names
        new_contig1 = [x for x in self.names if x != contig][0]
        new_contig2 = [x for x in other_edge.names if x != contig][0]
        assert new_contig1 != new_contig2
        coords_intersection = self.coords[contig].intersection(other_edge.coords[contig])
        if coords_intersection is None:
            return False
        assert self._change_hit_coords_with_intersection(contig, coords_intersection)
        assert other_edge._change_hit_coords_with_intersection(contig, coords_intersection)
        self._make_contig_forwards(contig)
        other_edge._make_contig_forwards(contig)
        if self.names[0] == new_contig1:
            index_to_change = 1
        elif self.names[1] == new_contig1:
            index_to_change = 0
        else:
            raise Error('Unexpected error in edge.merge_into(). Cannot continue')
        assert self.names[index_to_change] == contig
        self.names[index_to_change] = new_contig2
        self.rev[new_contig2] = other_edge.rev[new_contig2]
        self.coords[new_contig2] = other_edge.coords[new_contig2]
        del self.rev[contig]
        del self.coords[contig]
        return True




__all__ = [
    'assembly',
    'common',
    'contig',
    'contig_trim',
    'egg_extract',
    'edge',
    'external_progs',
    'graph',
    'kcount',
    'kmers',
    'kraken',
    'mapping',
    'mummer',
    'qc',
    'qc_external',
    'read_trim',
    'seed',
    'seed_processor',
]
from iva import *



import os
import sys
import tempfile
import shutil
import pyfastaq
import pysam
from iva import common, mapping

class Error (Exception): pass

def _head_fastaq(reads1, reads2, outfile, count):
    '''Takes first N sequences from a pair of interleaved fasta/q files. Output is in FASTA format. Returns hash of read length distribution (key=read length, value=count)'''
    seq_reader1 = pyfastaq.sequences.file_reader(reads1)
    if reads2 is not None:
        seq_reader2 = pyfastaq.sequences.file_reader(reads2)
    f = pyfastaq.utils.open_file_write(outfile)
    lengths = {}
    original_line_length = pyfastaq.sequences.Fasta.line_length
    pyfastaq.sequences.Fasta.line_length = 0
    i = 0

    for seq1 in seq_reader1:
        if reads2 is not None:
            seq2 = next(seq_reader2)
        else:
            seq2 = None
        for seq in (seq1, seq2):
            if seq is None:
                continue
            lengths[len(seq)] = lengths.get(len(seq), 0) + 1
            if type(seq) == pyfastaq.sequences.Fastq:
                print(pyfastaq.sequences.Fasta(seq.id, seq.seq), file=f)
            else:
                print(seq, file=f)
            i += 1
        if i >= count:
            break

    pyfastaq.utils.close(f)
    pyfastaq.sequences.Fasta.line_length = original_line_length
    return lengths


def _median(d):
    '''Returns the median key from histogram (as a dict with values=counts) of frequencies'''
    assert(len(d))
    count = 0
    total = sum(d.values())
    for key in sorted(d.keys()):
        count += d[key]
        if count >= 0.5 * total:
            return key


def _run_kmc_with_script(script, reads, outfile, kmer, min_count, max_count, m_option, verbose, allow_fail):
    f = pyfastaq.utils.open_file_write(script)
    print('set -e', file=f)
    kmc_command = ''.join([
        'kmc -fa',
         ' -m', str(m_option),
         ' -k', str(kmer),
         ' -sf', '1',
         ' -ci', str(min_count),
         ' -cs', str(max_count),
         ' -cx', str(max_count),
         ' ', reads,
         ' kmc_out',
         ' $PWD'
    ])
    print(kmc_command, end='', file=f)
    if verbose >= 2:
        print('', file=f)
        print('run kmc:', os.getcwd(), kmc_command)
    else:
        print(' > /dev/null', file=f)

    print('kmc_dump', 'kmc_out', 'kmc_out.dump', file=f)
    print('sort -k2nr', 'kmc_out.dump >', outfile, file=f)
    pyfastaq.utils.close(f)
    return common.syscall('bash ' + script, allow_fail=allow_fail)


def _run_kmc(reads, outprefix, kmer, min_count, max_count, verbose=0):
    '''Runs the kmer counting program kmc on a FASTA file. Returns filename made by kmc of the counts of kmers'''
    reads = os.path.abspath(reads)
    tmpdir = tempfile.mkdtemp(prefix='tmp.run_kmc.', dir=os.getcwd())
    original_dir = os.getcwd()
    kmer_counts_file = os.path.abspath(outprefix + '.kmer_counts')
    os.chdir(tmpdir)

    # KMC seems a bit flaky with the -m for RAM option.and dies striaght away.
    # The range is 4-32 (GB).
    # Try 4 and 32 (the default), then give up. This seems to make a difference, regardless of
    # RAM available on the machine.
    ran_ok = _run_kmc_with_script('run_kmc.sh', reads, kmer_counts_file, kmer, min_count, max_count, 32, verbose, True)
    if not ran_ok:
        if verbose:
            print('First try of running kmc failed. Trying again with -m4 instead of -m32...', flush=True)
        ran_ok = _run_kmc_with_script('run_kmc.sh', reads, kmer_counts_file, kmer, min_count, max_count, 4, verbose, False)

    os.chdir(original_dir)
    shutil.rmtree(tmpdir)

    if not ran_ok:
        raise Error('Error running kmc. Cannot continue')

    return kmer_counts_file


def _kmc_to_kmer_counts(infile, number, kmers_to_ignore=None, contigs_to_check=None, verbose=0, threads=1):
    '''Makes a dict of the most common kmers from the kmer counts output file of kmc'''
    counts = {}
    if os.path.getsize(infile) == 0:
        return counts
    tmpdir = tempfile.mkdtemp(prefix='tmp.common_kmers.', dir=os.getcwd())
    ref_seqs_file = os.path.join(tmpdir, 'ref.fa')
    counts_fasta_file = os.path.join(tmpdir, 'counts.fa')
    using_refs = _write_ref_seqs_to_be_checked(ref_seqs_file, kmers_to_ignore=kmers_to_ignore, contigs_to_check=contigs_to_check)

    if not using_refs:
        if verbose > 2:
            print('No existing kmers or contigs to check against. Using most common kmer for seed', flush=True)
        f = pyfastaq.utils.open_file_read(infile)
        for line in f:
            if len(counts) >= number:
                break
            try:
                kmer, count = line.rstrip().split()
                count = int(count)
            except:
                raise Error('Error getting kmer info from this line:\n' + line)

            counts[kmer] = count
        pyfastaq.utils.close(f)
    else:
        if verbose > 2:
            print('Existing kmers or contigs to check against. Running mapping', flush=True)
        mapping_prefix = os.path.join(tmpdir, 'map')
        bam = mapping_prefix + '.bam'
        _counts_file_to_fasta(infile, counts_fasta_file)
        mapping.map_reads(counts_fasta_file, None, ref_seqs_file, mapping_prefix, minid=0.9, index_k=9, index_s=1, sort=False, verbose=verbose, required_flag='0x4', threads=threads)

        sam_reader = pysam.Samfile(bam, "rb")
        for sam in sam_reader.fetch(until_eof=True):
            if len(counts) >= number:
                break
            try:
                count = sam.qname.split('_')[1]
            except:
                raise Error('Error getting count from sequence name in bam:\n' + sam.qname)

            counts[common.decode(sam.seq)] = count
        sam_reader.close()

    shutil.rmtree(tmpdir)
    return counts


def _write_ref_seqs_to_be_checked(outfile, kmers_to_ignore=None, contigs_to_check=None):
    if (kmers_to_ignore is None or len(kmers_to_ignore) == 0) and (contigs_to_check is None or len(contigs_to_check) == 0):
        return False

    f = pyfastaq.utils.open_file_write(outfile)
    i = 1

    if kmers_to_ignore is not None:
        for kmer in kmers_to_ignore:
            print('>', i, sep='', file=f)
            print(kmer, file=f)
            i += 1

    if contigs_to_check is not None and len(contigs_to_check) > 0:
        original_line_length = pyfastaq.sequences.Fasta.line_length
        pyfastaq.sequences.Fasta.line_length = 0
        for name in contigs_to_check:
            if len(contigs_to_check[name].fa) > 20:
                print(contigs_to_check[name].fa, file=f)
        pyfastaq.sequences.Fasta.line_length = original_line_length

    pyfastaq.utils.close(f)
    return True


def _counts_file_to_fasta(infile, outfile):
    fin = pyfastaq.utils.open_file_read(infile)
    fout = pyfastaq.utils.open_file_write(outfile)
    i = 1
    for line in fin:
        try:
            kmer, count = line.rstrip().split()
            count = int(count)
        except:
            raise Error('Error getting kmer info from this line:\n' + line)

        print('>', i, '_', count, sep='', file=fout)
        print(kmer, file=fout)
        i += 1

    pyfastaq.utils.close(fin)
    pyfastaq.utils.close(fout)


def get_most_common_kmers(reads1, reads2, kmer_length=None, head=100000, min_count=10, max_count=100000000, most_common=100, method='kmc', verbose=0, ignore_seqs=None, contigs_to_check=None, threads=1):
    '''Gets the most common kmers from a pair of interleaved read FASTA or FASTQ files. Takes the first N sequences (determined by head).  Returns a dict of kmer=>frequency. If kmer length is not given, use min(0.8 * median read length, 95)'''
    tmpdir = tempfile.mkdtemp(prefix='tmp.common_kmers.', dir=os.getcwd())
    counts = {}
    reads = os.path.join(tmpdir, 'reads.fa')
    read_lengths = _head_fastaq(reads1, reads2, reads, head)
    if len(read_lengths) == 0:
        shutil.rmtree(tmpdir)
        return counts
    if kmer_length is None:
        kmer_length = min(int(0.8 * _median(read_lengths)), 95)

    if method == 'kmc':
        counts_file = _run_kmc(reads, os.path.join(tmpdir, 'out'), kmer_length, min_count, max_count, verbose=verbose)
        counts = _kmc_to_kmer_counts(counts_file, most_common, kmers_to_ignore=ignore_seqs, contigs_to_check=contigs_to_check, verbose=verbose, threads=threads)
    else:
        raise Error('Method "' + method + '" not supported in kcount.get_most_common_kmers(). Cannot continue.')

    shutil.rmtree(tmpdir)
    return counts



import shutil
import subprocess
import re
import sys
import pyfastaq
from iva import common

class Error (Exception): pass


def is_in_path(prog):
    return shutil.which(prog) is not None


prog_to_version_cmd = {
    'blastn': ('blastn -version', re.compile('^blastn: (.*)$')),
    'makeblastdb': ('makeblastdb -version', re.compile('makeblastdb: (.*)$')),
    'kmc': ('kmc', re.compile('^K-Mer Counter \(KMC\) ver\. (.*) \(.*\)$')),
    'kmc_dump': ('kmc_dump', re.compile('^KMC dump ver. (.*) \(.*\)$')),
    'kraken': ('kraken --version', re.compile('^Kraken version (.*)$')),
    'kraken-build': ('kraken-build --version', re.compile('^Kraken version (.*)$')),
    'nucmer': ('nucmer --version', re.compile('^NUCmer \(NUCleotide MUMmer\) version (.*)$')),
    'R': ('R --version', re.compile('^R version (.*) \(.*\) --')),
    'smalt': ('smalt version', re.compile('^Version: (.*)$')),
    'samtools': ('samtools', re.compile('^Version: (.*)$')),
}


assembly_progs = [
    'kmc',
    'kmc_dump',
    'nucmer',
    'smalt',
    'samtools',
]


qc_progs = [
    'nucmer',
    'R',
    'smalt',
    'samtools',
]


qc_progs_optional = [
    'blastn',
    'makeblastdb',
    'kraken',
    'kraken-build',
]


qc_make_db_progs = [
    'kraken',
    'kraken-build',
]


def get_version(prog, must_be_in_path=True):
    assert prog in prog_to_version_cmd
    if not is_in_path(prog):
        if must_be_in_path:
            raise Error('Error getting version of ' + prog + ' - not found in path.')
        else:
            return 'UNKNOWN - not in path'

    cmd, regex = prog_to_version_cmd[prog]
    cmd_output = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
    cmd_output = common.decode(cmd_output[0]).split('\n')[:-1] + common.decode(cmd_output[1]).split('\n')[:-1]
    for line in cmd_output:
        hits = regex.search(line)
        if hits:
            return hits.group(1)
    return 'UNKNOWN ...\n    I tried running this to get the version: "' + cmd + '"\n    and the output didn\'t match this regular expression: "' + regex.pattern + '"'


def get_all_versions(progs, must_be_in_path=True):
    info = []
    for prog in sorted(progs):
        version = get_version(prog, must_be_in_path=must_be_in_path)
        info.append(' '.join(['Using', prog, 'version', version]))
    return info


def write_prog_info(script, filename):
    if script == 'iva':
        required = assembly_progs
        optional = None
    elif script == 'iva_qc':
        required = qc_progs
        optional = qc_progs_optional
    elif script == 'iva_qc_make_db':
        required = qc_make_db_progs
        optional = None
    else:
        raise Error('Script ' + script + ' not recognised')

    f = pyfastaq.utils.open_file_write(filename)
    print(' '.join(sys.argv), file=f)
    print('IVA version', common.version, file=f)
    print('\n'.join(get_all_versions(required)), file=f)
    if optional is not None:
        print('\n'.join(get_all_versions(optional, must_be_in_path=False)), file=f)
    pyfastaq.utils.close(f)




import subprocess
import tempfile
import shutil
import os
import sys
import inspect
import pyfastaq
import iva

class Error (Exception): pass

gage_stats = [
    'Missing Reference Bases',
    'Missing Assembly Bases',
    'Missing Assembly Contigs',
    'Duplicated Reference Bases',
    'Compressed Reference Bases',
    'Bad Trim',
    'Avg Idy',
    'SNPs',
    'Indels < 5bp',
    'Indels >= 5',
    'Inversions',
    'Relocation',
    'Translocation',
]


ratt_stats = [
     'elements_found',
     'elements_transferred',
     'elements_transferred_partially',
     'elements_split',
     'parts_of_elements_not_transferred',
     'elements_not_transferred',
     'gene_models_to_transfer',
     'gene_models_transferred',
     'gene_models_transferred_partially',
     'exons_not_transferred_from_partial_matches',
     'gene_models_not_transferred',
]


def dummy_gage_stats():
    return {x:'NA' for x in gage_stats}


def dummy_ratt_stats():
    return {x:'NA' for x in ratt_stats}


def run_gage(reference, scaffolds, outdir, nucmer_minid=80, clean=True):
    reference = os.path.abspath(reference)
    scaffolds = os.path.abspath(scaffolds)
    ref = 'ref.fa'
    scaffs = 'scaffolds.fa'
    contigs = 'contigs.fa'
    gage_out = 'gage.out'
    gage_script = 'run.sh'
    cwd = os.getcwd()
    os.mkdir(outdir)
    os.chdir(outdir)
    extractor = iva.egg_extract.Extractor(os.path.abspath(os.path.join(os.path.dirname(iva.__file__), os.pardir)))
    gage_code_indir = os.path.join('iva', 'gage')
    gage_code_outdir = 'gage_code'
    extractor.copy_dir(gage_code_indir, gage_code_outdir)
    os.symlink(reference, ref)
    os.symlink(scaffolds, scaffs)
    pyfastaq.tasks.scaffolds_to_contigs(scaffs, contigs, number_contigs=True)
    f = pyfastaq.utils.open_file_write(gage_script)
    print(' '.join([
        'sh',
        os.path.join(gage_code_outdir, 'getCorrectnessStats.sh'),
        ref,
        contigs,
        scaffs,
        str(nucmer_minid),
        '>', gage_out
        ]), file=f)
    pyfastaq.utils.close(f)
    iva.common.syscall('bash ' + gage_script, allow_fail=True)
    if not os.path.exists(gage_out):
        raise Error('Error running GAGE\nbash ' + gage_script)
    stats = dummy_gage_stats()
    wanted_stats = set(gage_stats)
    f = pyfastaq.utils.open_file_read(gage_out)
    got_all_stats = False

    for line in f:
        if line.startswith('Corrected Contig Stats'):
            got_all_stats = True
            break
        elif ':' in line:
            a = line.rstrip().split(': ')
            if a[0] in wanted_stats:
                stat = a[1]
                if '%' in stat:
                    stat = stat.split('(')[0]
                if stat.isdigit():
                    stats[a[0]] = int(stat)
                else:
                    stats[a[0]] = float(stat)
    pyfastaq.utils.close(f)

    if not got_all_stats:
        raise Error('Error running GAGE\nbash ' + gage_script)

    if clean:
        to_clean = [
            'contigs.fa.delta',
            'contigs.fa.fdelta',
            'contigs.fa.matches.lens',
            'out.1coords',
            'out.1delta',
            'out.mcoords',
            'out.mdelta',
            'out.qdiff',
            'out.rdiff',
            'out.snps',
            'out.unqry',
            'scaffolds.fa.coords',
            'scaffolds.fa.delta',
            'scaffolds.fa.err',
            'scaffolds.fa.fdelta',
            'scaffolds.fa.tiling',
            'tmp_scf.fasta',
        ]
        for f in to_clean:
            try:
                os.unlink(f)
            except:
                pass

    os.chdir(cwd)
    return stats


def run_ratt(embl_dir, assembly, outdir, config_file=None, transfer='Species', clean=True):
    embl_dir = os.path.abspath(embl_dir)
    assembly = os.path.abspath(assembly)

    cwd = os.getcwd()
    try:
        os.mkdir(outdir)
        os.chdir(outdir)
    except:
        raise Error('Error mkdir ' + outdir)

    extractor = iva.egg_extract.Extractor(os.path.abspath(os.path.join(os.path.dirname(iva.__file__), os.pardir)))
    ratt_code_indir = os.path.join('iva', 'ratt')
    ratt_code_outdir = 'ratt_code'
    extractor.copy_dir(ratt_code_indir, ratt_code_outdir)

    if config_file is None:
        ratt_config = os.path.join(ratt_code_outdir, 'ratt.config')
    else:
        ratt_config = os.path.abspath(config_file)

    script = 'run.sh'
    script_out = 'run.sh.out'
    ratt_outprefix = 'out'
    f = pyfastaq.utils.open_file_write(script)
    print('export RATT_HOME=', ratt_code_outdir, sep='', file=f)
    print('export RATT_CONFIG=', ratt_config, sep='', file=f)
    print('$RATT_HOME/start.ratt.sh', embl_dir, assembly, ratt_outprefix, transfer, file=f)
    pyfastaq.utils.close(f)
    cmd = 'bash ' + script + ' > ' + script_out
    # sometimes ratt returns nonzero code, but is OK, so ignore it
    iva.common.syscall(cmd, allow_fail=True)

    stats = {}

    matches = {
        'elements found.': 'elements_found',
        'Elements were transfered.': 'elements_transferred',
        'Elements could be transfered partially.': 'elements_transferred_partially',
        'Elements split.': 'elements_split',
        'Parts of elements (i.e.exons tRNA) not transferred.': 'parts_of_elements_not_transferred',
        'Elements couldn\'t be transferred.': 'elements_not_transferred',
        'Gene models to transfer.': 'gene_models_to_transfer',
        'Gene models transferred correctly.': 'gene_models_transferred',
        'Gene models partially transferred.': 'gene_models_transferred_partially',
        'Exons not transferred from partial CDS matches.': 'exons_not_transferred_from_partial_matches',
        'Gene models not transferred.': 'gene_models_not_transferred',
    }

    f = pyfastaq.utils.open_file_read(script_out)
    for line in f:
        if '\t' in line:
            a = line.rstrip().split('\t')
            if len(a) == 2 and a[0].isdigit() and a[1] in matches:
                stats[matches[a[1]]] = int(a[0])
    pyfastaq.utils.close(f)

    if clean:
        for d in ['Query', 'Reference', 'Sequences']:
            try:
                shutil.rmtree(d)
            except:
                pass

        iva.common.syscall('rm query.* Reference.* nucmer.* out.*')

    os.chdir(cwd)
    return stats


def run_blastn_and_write_act_script(assembly, reference, blast_out, script_out):
    tmpdir = tempfile.mkdtemp(prefix='tmp.blastn.', dir=os.getcwd())
    assembly_union = os.path.join(tmpdir, 'assembly.union.fa')
    reference_union = os.path.join(tmpdir, 'reference.union.fa')
    pyfastaq.tasks.to_fasta_union(assembly, assembly_union, seqname='assembly_union')
    pyfastaq.tasks.to_fasta_union(reference, reference_union, seqname='reference_union')
    iva.common.syscall('makeblastdb -dbtype nucl -in ' + reference_union)
    cmd = ' '.join([
        'blastn',
        '-task blastn',
        '-db', reference_union,
        '-query', assembly_union,
        '-out', blast_out,
        '-outfmt 6',
        '-evalue 0.01',
        '-dust no',
    ])
    iva.common.syscall(cmd)

    f = pyfastaq.utils.open_file_write(script_out)
    print('#!/usr/bin/env bash', file=f)
    print('act', reference, blast_out, assembly, file=f)
    pyfastaq.utils.close(f)
    iva.common.syscall('chmod 755 ' + script_out)
    shutil.rmtree(tmpdir)



import inspect
import sys
import os
import tempfile
import shutil
import time
import re
import urllib.request
import iva
import pyfastaq


class Error (Exception): pass


class Database:
    def __init__(self, rootdir, extra_refs_file=None, threads=1, minimizer_len=13, max_db_size=3, preload=False, verbose=False, skip_virus_download=False):
        self.rootdir = os.path.abspath(rootdir)
        if extra_refs_file is None:
            self.extra_refs_file = None
        else:
            self.extra_refs_file = os.path.abspath(extra_refs_file)

        self.threads = threads
        self.minimizer_len = minimizer_len
        self.max_db_size = max_db_size
        self.current_taxon_id = 2000000000
        self.current_gi = 4000000000 
        self.preload = preload
        self.verbose = verbose
        self.taxon_to_parent = {}
        self.extra_refs = {} # keys =  new fake taxon ID
        self.reseq_virus_dir = os.path.join(self.rootdir, 'Virus_refseq')
        self.kraken_db = os.path.join(self.rootdir, 'Kraken_db')
        self.kraken_taxon_dir = os.path.join(self.kraken_db, 'taxonomy')
        self.kraken_names_dmp = os.path.join(self.kraken_taxon_dir, 'names.dmp')
        self.kraken_nodes_dmp = os.path.join(self.kraken_taxon_dir, 'nodes.dmp')
        self.kraken_gi_taxid_nucl_dmp = os.path.join(self.kraken_taxon_dir, 'gi_taxid_nucl.dmp')
        self.embl_root = os.path.join(self.rootdir, 'EMBL')
        self.extra_refs_dir = os.path.join(self.rootdir, 'Extra_refs')
        self.skip_virus_download = skip_virus_download
        self.added_to_kraken = set()

        if self.skip_virus_download and extra_refs_file is None:
            raise Error('Cannot create database. skip_virus_download is True and no extra_refs_file provided')

        self.tasks = [
            'download',
            'build',
            'make_embl',
            'clean',
        ]

        self.done_files = {x:os.path.join(self.rootdir, 'progress.' + x + '.done') for x in self.tasks}


    def _mkdir(self, d, rmtree=False):
        if rmtree and os.path.exists(d):
            shutil.rmtree(d)

        if not os.path.exists(d):
            try:
                os.mkdir(d)
            except:
                raise Error('Error mkdir ' + d)


    def _get_parent_taxons(self, taxons):
        f = pyfastaq.utils.open_file_read(self.kraken_nodes_dmp)
        for line in f:
            a = line.split()
            if a[0] in taxons:
                self.taxon_to_parent[a[0]] = a[2]
        pyfastaq.utils.close(f)


    def _load_extra_ref_info(self):
        if self.extra_refs_file is None:
             return

        f = pyfastaq.utils.open_file_read(self.extra_refs_file)
        for line in f:
            genbank_ids = line.rstrip().split()
            new_gis = list(range(self.current_gi, self.current_gi + len(genbank_ids)))
            self.current_gi += len(genbank_ids)
            assert len(genbank_ids) == len(new_gis)
            self.extra_refs[self.current_taxon_id] = {
                'genbank_ids': genbank_ids,
                'new_gis': new_gis,
            }
            
            self.current_taxon_id += 1
        pyfastaq.utils.close(f)

        
    def _download_from_genbank(self, outfile, filetype, gi, max_tries=5, delay=3):
        assert filetype in ['gb', 'fasta']
        file_ok = False

        for i in range(max_tries):
            try:
                url = 'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=nucleotide&rettype=' + filetype + '&retmode=text&id=' + gi
                file_contents = iva.common.decode(urllib.request.urlopen(url).read()).rstrip()
            except:
                time.sleep(delay)
                continue

            if (filetype == 'fasta' and file_contents.startswith('>')) \
               or (filetype == 'gb' and file_contents.startswith('LOCUS')):
                file_ok = True
                break
            else:
                time.sleep(delay)

        if not file_ok:
            raise Error('Error downloading gi ' + gi + ' using:\n' + url + '\nI got this:\n' + file_contents)

        f = pyfastaq.utils.open_file_write(outfile)
        print(file_contents, file=f)
        pyfastaq.utils.close(f)


    def _download_extra_refs(self):
        if self.extra_refs_file is None:
             return

        for l in [x['genbank_ids'] for x in self.extra_refs.values()]:
            for gi in l:
                outprefix = os.path.join(self.extra_refs_dir, gi)
                self._download_from_genbank(outprefix + '.gb', 'gb', gi)
                self._download_from_genbank(outprefix + '.fasta', 'fasta', gi)


    def _genbank_to_taxon_and_gi(self, filename):
        f = pyfastaq.utils.open_file_read(filename)
        taxon_id = None
        gi = None
        for line in f:
            if line.startswith('                     /db_xref="taxon:'):
                taxon_id = line.rstrip().split(':')[-1].rstrip('"')            
            elif line.startswith('VERSION'):
                gi = line.rstrip().split()[-1].split(':')[-1]
            if None not in [taxon_id, gi]:
                break
        pyfastaq.utils.close(f)
        if None in [taxon_id, gi]:
            raise Error('Error getting taxon or GI from ' + filename)
        return taxon_id, gi


    def _genbank2embl(self, infile, outfile):
        tmpdir = tempfile.mkdtemp(prefix='tmp.genbank2embl.', dir=os.getcwd())
        extractor = iva.egg_extract.Extractor(os.path.abspath(os.path.join(os.path.dirname(iva.__file__), os.pardir)))
        genbank2embl_egg = os.path.join('iva', 'ratt', 'genbank2embl.pl')
        genbank2embl = os.path.join(tmpdir, 'genbank2embl.pl')
        extractor.copy_file(genbank2embl_egg, genbank2embl)
        os.chmod(genbank2embl, stat.S_IRWXU)

        # some genbank files have a 'CONTIG' line, which breaks bioperl's
        # conversion genbank --> embl and makes genbank2embl.pl hang
        iva.common.syscall('grep -v CONTIG ' + infile + ' > tmp.gbk; mv tmp.gbk ' + infile)
        iva.common.syscall(genbank2embl + ' ' + infile + ' ' + outfile, verbose=self.verbose)
        shutil.rmtree(tmpdir)
        

    def _append_to_file(self, filename, line):
        try:
            f = open(filename, 'a')
        except:
            raise Error('Error opening for appending:', filename)
        print(line, file=f)
        f.close()


    def _add_to_kraken(self, fa_file, real_taxon, new_taxon, new_gi):
        parent_taxon = self.taxon_to_parent[real_taxon]
        if self.verbose:
            print('add_to_kraken', fa_file, real_taxon, new_taxon, parent_taxon, new_gi, sep='\t')
        # update names.dmp: append new_taxon with whatever name that kraken will report
        # update nodes.dmp: append new_taxon with original parent taxon
        # update gi_taxid_nucl.dmp: append new_gi, new_taxon
        if new_taxon not in self.added_to_kraken:
            line = str(new_taxon) + '\t|\tadded.' + str(new_taxon) +  '\t|\t\t|\tscientific name\t|'
            self._append_to_file(self.kraken_names_dmp, line)
            line = '\t|\t'.join([
                str(new_taxon),
                parent_taxon,
                'species',
                'HI',
                '9',
                '1',
                '1',
                '1',
                '0',
                '1',
                '1',
                '0',
                '',
            ]) + '\t|'
            self._append_to_file(self.kraken_nodes_dmp, line)
        self.added_to_kraken.add(new_taxon)
        self._append_to_file(self.kraken_gi_taxid_nucl_dmp, str(new_gi) + '\t' + str(new_taxon)) 
        iva.common.syscall('kraken-build --add-to-library ' + fa_file + ' --db ' + self.kraken_db, verbose=self.verbose)


    def _replace_fasta_header(self, filename, header):
        fin = pyfastaq.utils.open_file_read(filename)
        tmp_out = filename + '.tmp.fa'
        fout = pyfastaq.utils.open_file_write(tmp_out)

        for line in fin:
            if line.startswith('>'):
                print('>' + header, file=fout)
            else:
                print(line.rstrip(), file=fout)

        pyfastaq.utils.close(fin)
        pyfastaq.utils.close(fout)
        os.rename(tmp_out, filename)


    def _sort_out_extra_refs(self):
        if self.extra_refs_file is None:
             return
        real_taxon_ids = set()

        for new_taxon_id, info in self.extra_refs.items():
            gb_file = os.path.join(self.extra_refs_dir, info['genbank_ids'][0] + '.gb')
            info['real_taxon'], info['gi'] = self._genbank_to_taxon_and_gi(gb_file)
            real_taxon_ids.add(info['real_taxon'])
            embl_dir = os.path.join(self.embl_root, 'added.' + str(new_taxon_id))
            self._mkdir(embl_dir)
            for i in range(len(info['genbank_ids'])):
                gi = info['genbank_ids'][i]
                new_gi = info['new_gis'][i]
                gb_file = os.path.join(self.extra_refs_dir, gi + '.gb')
                fa_file = os.path.join(self.extra_refs_dir, gi + '.fasta')
                self._replace_fasta_header(fa_file, 'gi|' + str(new_gi) + '|x')
                embl_file = os.path.join(embl_dir, gi + '.embl')
                self._genbank2embl(gb_file, embl_file)

        self._get_parent_taxons(real_taxon_ids)

        for new_taxon_id, info in  self.extra_refs.items():
            real_taxon = info['real_taxon']
            for i in  range(len(info['genbank_ids'])):
                gi = info['genbank_ids'][i]
                new_gi = info['new_gis'][i]
                fa_file = os.path.join(self.extra_refs_dir, gi + '.fasta')
                gb_file = os.path.join(self.extra_refs_dir, gi + '.gb')
                self._add_to_kraken(fa_file, real_taxon, new_taxon_id, new_gi)
                if self.verbose:
                    print('unlink', os.path.exists(gb_file), gb_file)
                    print('unlink', os.path.exists(fa_file), fa_file)
                os.unlink(gb_file)
                os.unlink(fa_file)
               

    def _build_kraken_virus_db(self):
        if os.path.exists(self.done_files['clean']):
            print('Kraken DB already built. Skipping.', flush=True)
            return

        if os.path.exists(self.done_files['download']):
            print('Files already downloaded. Skipping.', flush=True)
        else:
            for d in [self.rootdir, self.embl_root, self.extra_refs_dir]:
                self._mkdir(d, rmtree=True)

            iva.common.syscall('kraken-build --download-taxonomy --db ' + self.kraken_db, verbose=self.verbose)
            if not self.skip_virus_download:
                iva.common.syscall('kraken-build --download-library viruses --db ' + self.kraken_db, verbose=self.verbose)
        
            if self.extra_refs_file is not None:
                self._load_extra_ref_info()
                self._download_extra_refs()
                self._sort_out_extra_refs()
            iva.common.syscall('touch ' + self.done_files['download'], verbose=self.verbose)

        if os.path.exists(self.done_files['build']):
            print('Already built. Skipping.', flush=True)
        else:
            cmd = ' '.join([
                'kraken-build --build',
                '--db', self.kraken_db,
                '--max-db-size', str(self.max_db_size),
                '--minimizer-len', str(self.minimizer_len),
                '--threads', str(self.threads)
            ])
            iva.common.syscall(cmd, verbose=self.verbose)
            iva.common.syscall('touch ' + self.done_files['build'])


    def _clean(self):
        if os.path.exists(self.done_files['clean']):
            print('Already cleaned. Skipping.')
        else:
            iva.common.syscall('kraken-build --clean --db ' + self.kraken_db, verbose=self.verbose)
            if os.path.exists(self.extra_refs_dir):
                shutil.rmtree(self.extra_refs_dir)
            iva.common.syscall('touch ' + self.done_files['clean'], verbose=self.verbose)
             

    def _get_genbank_virus_files(self):
        if os.path.exists(self.done_files['make_embl']):
            print('Already made EMBL files. Skipping.')
            return
        self._mkdir(self.reseq_virus_dir, rmtree=True)
        cwd = os.getcwd()
        os.chdir(self.reseq_virus_dir)
        iva.common.syscall('wget -q ftp://ftp.ncbi.nlm.nih.gov/genomes/Viruses/all.gbk.tar.gz', verbose=self.verbose)
        iva.common.syscall('tar -xf all.gbk.tar.gz', verbose=self.verbose)
        os.chdir(cwd)
        self._convert_refseq_virus_to_embl()
        shutil.rmtree(self.reseq_virus_dir)
        iva.common.syscall('touch ' + self.done_files['make_embl'])


    def _convert_refseq_virus_to_embl(self):
        original_dir = os.getcwd()

        for directory, x, gbk_files in sorted(os.walk(self.reseq_virus_dir)):
            if directory == self.reseq_virus_dir or len(gbk_files) == 0:
                continue
            os.chdir(directory)
            if self.verbose:
                print('Converting', directory, end=' ', flush=True)
            for fname in gbk_files:
                self._genbank2embl(fname, re.sub('\.gbk$', '', fname) + '.embl')
                os.unlink(fname)
                if self.verbose:
                    print(fname, end=' ', flush=True)

            os.chdir(original_dir)
            if self.verbose:
                print()
            new_dir = re.sub('_uid[0-9]+$', '', directory).strip('_')
            if new_dir != directory:
                os.rename(directory, new_dir) 

            final_dir =  os.path.join(self.embl_root, os.path.basename(new_dir))
            if os.path.exists(final_dir):
                shutil.rmtree(final_dir)
            os.rename(new_dir, final_dir)

        os.chdir(original_dir)


    def build(self):
        self._build_kraken_virus_db()
        if not self.skip_virus_download:
            self._get_genbank_virus_files()
        self._clean()


    def _run_kraken(self, reads, outprefix):
        kraken_out = outprefix + '.out'
        kraken_report = outprefix + '.report'
        cmd = ' '.join([
            'kraken',
            '--db', self.kraken_db,
            '--preload' if self.preload else '',
            '--threads', str(self.threads),
            '--output', kraken_out,
            reads
        ])

        iva.common.syscall(cmd, verbose=self.verbose)

        cmd = ' '.join([
            'kraken-report',
            '--db', self.kraken_db,
            kraken_out,
            '>', kraken_report
        ])
        iva.common.syscall(cmd, verbose=self.verbose)
        os.unlink(kraken_out)


    def _species_to_embl_dir(self, s):
        if s.startswith('added.'):
            return s
        else:
            return re.sub('\W', '_', s).strip('_')


    def _get_most_common_species_dir(self, kraken_report):
        embl_dirs = set([os.path.basename(x[0]) for x in os.walk(self.embl_root) if x[0] != self.embl_root])
        f = pyfastaq.utils.open_file_read(kraken_report)
        max_count = -1
        most_common_dir = None
        for line in f:
            a = line.rstrip().split('\t')
            this_species = a[-1].strip()
            this_dir = self._species_to_embl_dir(this_species)
            if this_dir in embl_dirs and int(a[2]) > max_count:
                most_common_dir = this_dir
                max_count = int(a[2])

        pyfastaq.utils.close(f)
        if most_common_dir is None:
            return None
        else:
            return os.path.join(self.embl_root, most_common_dir)


    def choose_reference(self, reads, outprefix):
        self._run_kraken(reads, outprefix)
        lib_dir = os.path.join(self.rootdir, 'Library')
        return self._get_most_common_species_dir(outprefix + '.report')




from collections import Counter

class Kmers:
    def __init__(self, kmer=None, left=False, verbose=0):
        self.kmers = []
        self.left = left
        self.verbose = verbose
        if kmer is not None:
            self.append(kmer)


    def append(self, kmer):
        '''Adds one kmer'''
        if 'N' in kmer:
            return
        if self.left:
            self.kmers.append(kmer[::-1])
        else:
            self.kmers.append(kmer)


    def extend(self, l):
        '''Addds a list of kmers'''
        for x in l:
            self.append(x)


    def _kmer_dict(self, k):
        return dict(Counter([kmer[0:k] for kmer in self.kmers if len(kmer) >= k]))


    def _commonest_kmers(self, k):
        counts = self._kmer_dict(k)
        highest_counts = None, None
        highest_kmers = None, None

        for k in counts:
            if highest_counts[1] is None:
                highest_kmers = None, k
                highest_counts = None, counts[k]
            elif counts[k] > highest_counts[1]:
                highest_kmers = highest_kmers[1], k
                highest_counts = highest_counts[1], counts[k]
            elif highest_counts[0] is None:
                highest_kmers = k, highest_kmers[1]
                highest_counts = counts[k], highest_counts[1]
        return highest_kmers, highest_counts


    def extension(self, min_cov, min_ratio, extend_length):
        '''Returns a string that can extend the kmer, of length as long as possible up to extend_length bases. Must be supported by min_cov kmers, and also it must be at least min_ratio more abundant than the second most common kmer'''
        if self.verbose >= 4:
            print('        all kmers:')
            for kmer in self.kmers:
                print('            ', kmer)
        if len(self.kmers) == 0:
            return ''
        longest = max([len(x) for x in self.kmers])
        for i in range(min(longest, extend_length),0, -1):
            highest_kmers, highest_counts = self._commonest_kmers(i)
            if self.verbose >= 2:
                print('        k =', i, 'commonest two kmers:', highest_kmers, 'have frequency:', highest_counts)
            if highest_kmers[1] is None:
                pass
            elif highest_kmers[0] is None:
                if highest_counts[1] >= min_cov:
                    extension_seq = highest_kmers[1]
                    return highest_kmers[1][::-1] if self.left else highest_kmers[1]
            else:
                assert highest_counts[0] > 0 and highest_counts[1] > 0
                if highest_counts[1] / highest_counts[0] >= min_ratio and highest_counts[1] >= min_cov:
                    return highest_kmers[1][::-1] if self.left else highest_kmers[1]

        return ''
